#include <stdio.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>

#define FILENAME "NproducerMconsumer1buffer.c"
#define RANGE    16
#define TIME     30
#define P        4
#define C        4

#define PRELEVATO    0
#define DEPOSITATO    1

void releaseAll(int,int);

int main(int argc, char **argv) {
  int *buffer;
  int bufferid, semid;
  int i,j,M;
  int delay, time=TIME;
  int p=0;

  pid_t pid;
  pid_t cons_pid[C];

  struct sembuf sb;
  union semun {
    int val;                    /* value for SETVAL */
    struct semid_ds *buf;       /* buffer for IPC_STAT, IPC_SET */
    unsigned short int *array;  /* array for GETALL, SETALL */
    struct seminfo *__buf;      /* buffer for IPC_INFO */
  } su;


  sb.sem_flg=0;

  /* allocate shared memory segments */
  if ((bufferid = shmget(ftok(FILENAME,'B'),1,IPC_CREAT|0666)) == -1) {
    perror("shmget() for buffer");
    releaseAll(bufferid,semid);
    exit(-1);
  }

  /* allocate semaphores */
  if ((semid = semget(ftok(FILENAME,'S'),2,IPC_CREAT|0666)) == -1) {
    perror("semget()");
    releaseAll(bufferid,semid);
    exit(-4);
  }

  /* initialize semaphores and shared segments */
  su.val=0;
  if (semctl(semid,DEPOSITATO,SETVAL,su)==-1) {
    perror("semctl() to initialize DEPOSITATO");
    exit(-22);
  }
  su.val=1;
  if (semctl(semid,PRELEVATO,SETVAL,su)==-1) {
    perror("semctl() to initialize PRELEVATO");
    exit(-23);
  }

  /* fork producers */
  pid = -1;
  for(i=0; i<P && pid!=0; i++)
    pid=fork();

  switch(pid) {
  case -1:
    perror("fork()");
    releaseAll(bufferid,semid);
    exit(-5);
    break;
  case 0:
    /* GENERIC PRODUCER i */
    if ((int)(buffer=shmat(bufferid,NULL,0)) == -1) {
      perror("producer: shmat() for buffer");
      exit(-7);
    }
    /* PRODUCING */
    srandom(i);
    // produci un messaggio in M
    while (time>0) {
      sb.sem_num=PRELEVATO; sb.sem_op=-1;
      if (semop(semid, &sb,1)==-1) {  /* P(PRELEVATO) */
	perror("semop() producer P(PRELEVATO)");
	exit(-9);
      }
      M = (int)(random() % RANGE);

      buffer[0] = M;                 /* PRODUCI */

      printf("Producer %d produces: %d\n",i,M);

      sb.sem_num=DEPOSITATO; sb.sem_op= 1;
      if (semop(semid, &sb,1)==-1) {  /* V(DEPOSITATO) */
	perror("semop() producer V(DEPOSITATO)");
	exit(-12);
      }
      delay = (int)(random() % P ) / 2 + 1;
      sleep(delay);
      time -=delay;
    }
    printf("Producer %d exits\n",i);
    shmdt(buffer);
    return 0;
  }

  /* fork consumers */
  pid = -1;
  for (j=0; j<C && pid!=0; j++)
    pid = cons_pid[j] = fork();

  switch(pid) {
  case -1:
    perror("fork()");
    releaseAll(bufferid,semid);
    exit(-5);
    break;
  case 0:
    /* GENERIC CONSUMER j */
    if ((int)(buffer=shmat(bufferid,NULL,0)) == -1) {
      perror("consumer: shmat() fo buffer");
      exit(-7);
    }

    /* CONSUMING */
    srandom(j);
    // consuma un messaggio in M
    while (1) {
      sb.sem_num=DEPOSITATO; sb.sem_op=-1;
      if (semop(semid, &sb,1)==-1) {  /* P(DEPOSITATO) */
	perror("semop() consumer P(DEPOSITATO)");
	exit(-9);
      }

      M = buffer[0];                 /* CONSUMA */

      printf("Consumer %d consumes: %d\n",j,M);

      sb.sem_num=PRELEVATO; sb.sem_op= 1;
      if (semop(semid, &sb,1)==-1) {  /* V(PRELEVATO) */
	perror("semop() consumer V(PRELEVATO)");
	exit(-12);
      }
      delay = (int)(random() % C ) / 2 + 2;
      sleep(delay);
    }
    printf("Consumer %d exits\n",i);
    shmdt(buffer);
    return 0;
  }

  do {
    wait(NULL); p++;
  } while (p<P);  // wait producers ...

  sb.sem_num=PRELEVATO; sb.sem_op=-1;// wait that consumers consume everything ...
  if (semop(semid, &sb,1)==-1) {     // P(PRELEVATO)
    perror("semop() waiting for producers");
    exit(-19);
  }

  for(j=0; j<C; j++) {
    kill(cons_pid[j],SIGKILL); // ... and Kill consumers
    wait(NULL);
  }

  /* dealloca risorse */
  releaseAll(bufferid,semid);
  return 0;
}

void releaseAll(int bufferid,int semid) {
  shmctl(bufferid,IPC_RMID,0);
  semctl(semid,0,IPC_RMID,0);
}
