#include <stdio.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>

#define FILENAME "producerconsumer.c"
#define RANGE    16
#define TIME     30
#define N        5
#define P        4
#define C        4

#define VUOTE    0
#define PIENE    1
#define USO_T    2
#define USO_D    3

void releaseAll(int,int,int,int);
void printBuffer(int,int *,int);

int main(int argc, char **argv) {
  int *buffer,*T,*D;
  int bufferid, Tid, Did,semid;
  int i,j,M;
  int delay, time=TIME;
  int kids=P+C;
  int p=0;

  struct sembuf sb; 
  union semun {
    int val;                    /* value for SETVAL */
    struct semid_ds *buf;       /* buffer for IPC_STAT, IPC_SET */
    unsigned short int *array;  /* array for GETALL, SETALL */
    struct seminfo *__buf;      /* buffer for IPC_INFO */
  } su;
  pid_t pid;
  pid_t cons_pid[C];

  sb.sem_flg=0;

  /* allocate shared memory segments */
  if ((bufferid = shmget(ftok(FILENAME,'B'),N,IPC_CREAT|0666)) == -1) {
    perror("shmget() for buffer");
    releaseAll(bufferid,Tid,Did,semid);
    exit(-1);
  }
  if ((Tid = shmget(ftok(FILENAME,'D'),1,IPC_CREAT|0666)) == -1) {
    perror("shmget() for T");
    releaseAll(bufferid,Tid,Did,semid);
    exit(-2);
  }
  if ((Did = shmget(ftok(FILENAME,'T'),1,IPC_CREAT|0666)) == -1) {
    perror("shmget() for D");
    releaseAll(bufferid,Tid,Did,semid);
    exit(-3);
  }

  /* allocate semaphores */
  if ((semid = semget(ftok(FILENAME,'S'),5,IPC_CREAT|0666)) == -1) {
    perror("semget()");
    releaseAll(bufferid,Tid,Did,semid);
    exit(-4);
  }

  /* initialize semaphores and shared segments */
  if ((int)(T=shmat(Tid,NULL,0)) == -1) {
    perror("main: shmat() for T");
    exit(-5);
  }
  if ((int)(D=shmat(Did,NULL,0)) == -1) {
    perror("main: shmat() for D");
    exit(-6);
  }    
  *T = 0; *D = 0;
  shmdt(T); shmdt(D);

  su.val=1;
  if (semctl(semid,USO_D,SETVAL,su)==-1) {
    perror("semctl() to initialize USO_D");
    exit(-20);
  }
  if (semctl(semid,USO_T,SETVAL,su)==-1) {
    perror("semctl() to initialize USO_T");
    exit(-21);
  }
  su.val=0;
  if (semctl(semid,PIENE,SETVAL,su)==-1) {
    perror("semctl() to initialize PIENE");
    exit(-22);
  }
  su.val=N;
  if (semctl(semid,VUOTE,SETVAL,su)==-1) {
    perror("semctl() to initialize VUOTE");
    exit(-23);
  }
  pid = -1;
  for(i=0; i<P && pid!=0; i++)
    pid=fork();

  switch(pid) {
  case -1:
    perror("fork()");
    releaseAll(bufferid,Tid,Did,semid);
    exit(-5);
    break;
  case 0: 
    /* GENERIC PRODUCER i */
    if ((int)(buffer=shmat(bufferid,NULL,0)) == -1) {
      perror("producer: shmat() for buffer");
      exit(-7);
    }
    if ((int)(D=shmat(Did,NULL,0)) == -1) {
      perror("producer: shmat() for D");
      exit(-8);
    }
    if ((int)(T=shmat(Tid,NULL,0)) == -1) {
      perror("producer: shmat() for T");
      exit(-8);
    }
    /* PRODUCING */
    srandom(i);
    // produci un messaggio in M
    while (time>0) {
      sb.sem_num=VUOTE; sb.sem_op=-1;
      if (semop(semid, &sb,1)==-1) {  /* down(VUOTE) */ 
	perror("semop() producer down(VUOTE)");
	exit(-9);
      }
      sb.sem_num=USO_D; sb.sem_op=-1;
      if (semop(semid, &sb,1)==-1) {  /* down(USO_D) */ 
	perror("semop() producer down(USO_D)");
	exit(-10);
      }
      M = (int)(random() % RANGE);
      buffer[*D] = M;                 /* PRODUCI */
      *D = (*D + 1) % N;
      printf("Producer %d produces: %d\n",i,M);
      printBuffer(semid,buffer,*T);

      sb.sem_num=USO_D; sb.sem_op= 1;
      if (semop(semid, &sb,1)==-1) {  /* up(USO_D) */ 
	perror("semop() producer up(USO_D)");
	exit(-11);
      }
      sb.sem_num=PIENE; sb.sem_op= 1;
      if (semop(semid, &sb,1)==-1) {  /* up(PIENE) */ 
	perror("semop() producer up(PIENE)");
	exit(-12);
      }
      delay = (int)(random() % P ) / 2 + 1;
      sleep(delay);
      time -=delay;
    }
    printf("Producer %d exits\n",i);
    shmdt(buffer);
    shmdt(D);
    return 0;
  }

  pid = -1;
  for (j=0; j<C && pid!=0; j++)
    pid = cons_pid[j] = fork();

  switch(pid) {
  case -1: 
    perror("fork()");
    releaseAll(bufferid,Tid,Did,semid);
    exit(-5);
    break;
  case 0: 
    /* GENERIC CONSUMER j */
    if ((int)(buffer=shmat(bufferid,NULL,0)) == -1) {
      perror("consumer: shmat() fo buffer");
      exit(-7);
    }
    if ((int)(T=shmat(Tid,NULL,0)) == -1) {
      perror("consumer: shmat() for T");
      exit(-8);
    }

    /* CONSUMING */
    srandom(j);
    // consuma un messaggio in M
    while (1) {
      sb.sem_num=PIENE; sb.sem_op=-1;
      if (semop(semid, &sb,1)==-1) {  /* down(PIENE) */ 
	perror("semop() consumer down(PIENE)");
	exit(-9);
      }
      sb.sem_num=USO_T; sb.sem_op=-1;
      if (semop(semid, &sb,1)==-1) {  /* down(USO_T) */ 
	perror("semop() consumer down(USO_T)");
	exit(-10);
      }

      M = buffer[*T];                 /* CONSUMA */
      *T = (*T + 1) % N;
      printf("Consumer %d consumes: %d\n",j,M);
      printBuffer(semid,buffer,*T);

      sb.sem_num=USO_T; sb.sem_op= 1;
      if (semop(semid, &sb,1)==-1) {  /* up(USO_T) */ 
	perror("semop() consumer up(USO_T)");
	exit(-11);
      }
      sb.sem_num=VUOTE; sb.sem_op= 1;
      if (semop(semid, &sb,1)==-1) {  /* up(VUOTE) */ 
	perror("semop() consumer up(VUOTE)");
	exit(-12);
      }
      delay = (int)(random() % C ) / 2 + 2;
      sleep(delay);
    }
    printf("Consumer %d exits\n",i);
    shmdt(buffer);
    shmdt(T);
    return 0;
  }

  do {
    wait(NULL); p++;
  } while (p<P);  // wait producers ...
  
  sb.sem_num=VUOTE; sb.sem_op=-N;// wait that consumers consume everything ...
  if (semop(semid, &sb,1)==-1) { 
    perror("semop() waiting for producers");
    exit(-19);
  }
  
  for(j=0; j<C; j++) {
    kill(cons_pid[j],SIGKILL); // ... and Kill consumers
    wait(NULL);
  }

  /* dealloca risorse */
  releaseAll(bufferid,Tid,Did,semid);
  return 0;
}

void releaseAll(int bufferid,int Tid, int Did, int semid) {
  shmctl(bufferid,IPC_RMID,0);
  shmctl(Tid,IPC_RMID,0);
  shmctl(Did,IPC_RMID,0);
  semctl(semid,0,IPC_RMID,0);
}

void printBuffer(int semid, int* buffer, int T) {
  // printf("T: %d\tD: %d\n",T,D);
  int i;
  int l = semctl(semid,PIENE,GETVAL,NULL);
  printf(">");
  if (l>0) printf("%d",buffer[T]);
  for(i = 1; i<l; i++)
    printf(" %d",buffer[(T+i) % N]);
  printf("<\n");
}
